#ifndef _evolutionaryAlgorithm_h
#define _evolutionaryAlgorithm_h

#include <iostream>
#include <vector>
#include <init.h>


/******************************************************
 * class to implement an genetic algorithm
 */
class EvolutionaryAlgorithm {
public:
  // constructor with all parameters of the GA
  EvolutionaryAlgorithm(Eval & _evalFunc, 
			int _mu, int _lambda, 
			Init & _initSol, 
			int _tournamentSize, 
			double _crossOverRate, double _mutationRate, 
			int _maxGeneration) : eval(_evalFunc), initSol(_initSol) { 
    mu = _mu;
    lambda = _lambda;
    tournamentSize = _tournamentSize;
    crossOverRate = _crossOverRate;
    mutationRate = _mutationRate ;
    maxGeneration = _maxGeneration;
  }

  /*
   *
   * Evaluation of the population
   *
   **/
  void evalPop(Population & population) {
    for(int i = 0; i < population.size(); i++) {
      eval(population[i]);
    }
  }

  /*
   *
   * Initialization operator
   *
   **/
  void initialization(Population & parents) {
    parents.resize(mu);
    for(int i = 0; i < parents.size(); i++) {
      initSol(parents[i]);
    }
  }

  Solution * tournament(Population & parents) {
    Solution * winer = & parents[ random() % parents.size() ];
    Solution * oneGuy = NULL;
    for(int i = 1; i < tournamentSize; i++) {
      oneGuy = & parents[ random() % parents.size() ];
      if (oneGuy->fitness() > winer->fitness()) 
	winer = oneGuy;
    }

    return winer;
  }

  /*
   *
   * Selection operator
   *
   **/
  void selection(Population & parents, Population & genitors) {
    genitors.resize(lambda);
    for(int i = 0; i < genitors.size(); i++) {
      genitors[i] = *(tournament(parents));
    }	
  }

  /*
   * Crossover operator : uniform crossover
   *
   */
  void crossOver(Solution & sol1, Solution & sol2) {
    for(int i = 0; i < sol1.size(); i++) {
      if (random() / (double) RAND_MAX < 0.5) {
	bool tmp = sol1[i];
	sol1[i] = sol2[i];
	sol2[i] = tmp;
      }
    }
  }

  void mutation(Solution & sol) {
    double rate = ((double) 1) / (double) sol.size();
    for(int i = 0; i < sol.size(); i++) {
      if (random() / (double) RAND_MAX < rate) 
	sol[i] = ! sol[i];
    }
  }

  /*
   *
   * random variation operator : crossover and mutation operators
   *
   **/
  void randomVariation(Population & genitors) {
    for(int i = 0; i < genitors.size(); i += 2) {
      if (random() / (double) RAND_MAX < crossOverRate) 
	crossOver(genitors[i], genitors[i+1]);

      if (random() / (double) RAND_MAX < mutationRate)
	mutation(genitors[i]);

      if (random() / (double) RAND_MAX < mutationRate)
	mutation(genitors[i+1]);
    }
  }

  /*
   *
   * Replacement operator : (lambda + mu)
   * the mu best solutions from parents + children
   *
   **/
  void replacement(Population & parents, Population & children) {
    // take the space for parents + children
    parents.reserve(parents.size() + children.size());

    // put the children at the following of the parents
    for (int i = 0; i < children.size(); i++) {
      parents.push_back(children[i]);
    }

    // sort parents 
    parents.sort();

    // reduce the size to mu
    parents.resize(mu);
  }

  // method to perform the search
  void run(Population & parents) {
    int nbGeneration = 0;

    Population children;

    initialization(parents);
    evalPop(parents);

    // only for testing : print
    std::cout << nbGeneration << " " ;
    parents.bestSolution().print();
    std::cout << std::endl;
    //    parents.print();
    //    std::cout << std::endl;

    while (nbGeneration < maxGeneration) {
      selection(parents, children);
      randomVariation(children);
      evalPop(children);
      replacement(parents, children);
      nbGeneration += 1;

      // only for testing : print
      std::cout << nbGeneration << " " ;
      parents.bestSolution().print();
      std::cout << std::endl;
      //      parents.print();
      //      std::cout << std::endl;
    }
  }

protected:
  // evaluation function
  Eval & eval;

  // population size of the parents
  int mu;

  // population size of childrens
  int lambda;

  // initialization method of one solution
  Init & initSol;

  // size of the tournament for selection
  int tournamentSize;

  // crossover rate
  double crossOverRate ;

  // mutation rate
  double mutationRate;

  // stopping criterium : maximum number of generation
  int maxGeneration;

};

#endif
