import java.util.Random;
import java.util.Arrays;
import java.util.Comparator;

/******************************************************
 * class to implement an genetic algorithm
 */
public class EvolutionaryAlgorithm {
    // constructor with all parameters of the GA
    EvolutionaryAlgorithm(Random random, 
			  Eval evalFunc, 
			  int mu, int lambda, 
			  Init initSol, 
			  int tournamentSize, 
			  double crossOverRate, double mutationRate, 
			  int maxGeneration) { 
	this.random = random;
	this.eval = evalFunc;
	this.mu = mu;
	this.lambda = lambda;
	this.initSol = initSol;
	this.tournamentSize = tournamentSize;
	this.crossOverRate = crossOverRate;
	this.mutationRate = mutationRate ;
	this.maxGeneration = maxGeneration;

	this.nbGeneration = 0;

	this.comp = new SolutionComparator();
    }

    /*
     *
     * Evaluation of the population
     *
     **/
    void evalPop(Population population) {
	for(int i = 0; i < population.getSize(); i++) {
	    this.eval.apply(population.individuals[i]);
	}
    }

    /*
     *
     * Initialization operator
     *
     **/
    void initialization(Population parents) {
	parents.resize(this.mu);
	for(int i = 0; i < parents.getSize(); i++) {
	    this.initSol.init(parents.individuals[i]);
	}
    }

    Solution tournament(Population parents) {
	Solution winer = parents.individuals[ random.nextInt(parents.getSize()) ];
	Solution oneGuy;
	for(int i = 1; i < tournamentSize; i++) {
	    oneGuy = parents.individuals[ random.nextInt(parents.getSize()) ];
	    if (oneGuy.fitness() > winer.fitness()) 
		winer = oneGuy;
	}

	// remember to copy (duplicate) this solution
	return winer.clone();
    }

    /*
     *
     * Selection operator
     *
     **/
    void selection(Population parents, Population genitors) {
	for(int i = 0; i < genitors.getSize(); i++) {
	    genitors.individuals[i] = tournament(parents);
	}	
    }

    /*
     * Crossover operator : uniform crossover
     *
     */
    void crossOver(Solution sol1, Solution sol2) {
	for(int i = 0; i < sol1.size(); i++) {
	    if (random.nextFloat() < 0.5) {
		boolean tmp = sol1.bitString[i];
		sol1.bitString[i] = sol2.bitString[i];
		sol2.bitString[i] = tmp;
	    }
	}
    }

    void mutation(Solution sol) {
	double rate = ((double) 1) / (double) sol.size();
	for(int i = 0; i < sol.size(); i++) {
	    if (random.nextFloat() < rate) 
		sol.bitString[i] = ! sol.bitString[i];
	}
    }

    /*
     *
     * random variation operator : crossover and mutation operators
     *
     **/
    void randomVariation(Population genitors) {
	for(int i = 0; i < genitors.getSize(); i += 2) {
	    if (random.nextFloat() < crossOverRate) 
		crossOver(genitors.individuals[i], genitors.individuals[i+1]);

	    if (random.nextFloat() < mutationRate)
		mutation(genitors.individuals[i]);

	    if (random.nextFloat() < mutationRate)
		mutation(genitors.individuals[i+1]);
	}
    }

    /*
     *
     * Replacement operator : (lambda + mu)
     * the mu best solutions from parents + children
     *
     **/
    void replacement(Population parents, Population children) {
	// sort parents and children
	Arrays.sort(parents.individuals, comp);
	Arrays.sort(children.individuals, comp);

	// take the mu best solutions
	Solution [] newIndiv = new Solution[mu];

	int iparents = 0;
	int ichildren = 0;

	for(int i = 0; i < mu; i++) {
	    if (ichildren < children.getSize() && parents.individuals[iparents].fitness() < children.individuals[ichildren].fitness()) {
		newIndiv[i] = children.individuals[ichildren];
		ichildren += 1;
	    } else {
		newIndiv[i] = parents.individuals[iparents];
		iparents += 1;
	    }
	}

	parents.individuals = newIndiv;
    }

    // method to perform the search
    void run(Population parents) {
	this.nbGeneration = 0;

	Population children = new Population(lambda);

	initialization(parents);
	evalPop(parents);

	// only for testing : print
	System.out.println(nbGeneration + " " + parents.bestSolution());

	while (this.nbGeneration < this.maxGeneration) {
	    selection(parents, children);
	    randomVariation(children);
	    evalPop(children);
	    replacement(parents, children);
	    this.nbGeneration += 1;

	    // only for testing : print
	    System.out.println(nbGeneration + " " + parents.bestSolution());
	}
    }

    // number of generations
    int nbGeneration() {
	return this.nbGeneration;
    }

    // random generator
    protected Random random;

    // evaluation function
    protected Eval eval;

    // population size of the parents
    protected int mu;

    // population size of childrens
    protected int lambda;

    // initialization method of one solution
    protected Init initSol;

    // size of the tournament for selection
    protected int tournamentSize;

    // crossover rate
    protected double crossOverRate ;

    // mutation rate
    protected double mutationRate;

    // stopping criterium : maximum number of generation
    protected int maxGeneration;

    // number of generation
    private int nbGeneration;

    // comparator of solutions (used in replacement)
    private Comparator<Solution> comp;
}
